#ifndef DRACO_IO_FILE_READER_TEST_COMMON_H_
#define DRACO_IO_FILE_READER_TEST_COMMON_H_

#include <cstddef>

namespace draco {

const size_t kFileSizeCarDrc = 69892;
const size_t kFileSizeCubePcDrc = 224;

}  // namespace draco

#endif  // DRACO_IO_FILE_READER_TEST_COMMON_H_
